#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "target.h"
#include "tree.h"
#include "tree-iterator.h"
#include "input.h"
#include "diagnostic.h"
#include "stringpool.h"
#include "cgraph.h"
#include "gimplify.h"
#include "gimple-expr.h"
#include "convert.h"
#include "print-tree.h"
#include "stor-layout.h"
#include "fold-const.h"
#include "langhooks.h"
#include "langhooks-def.h"
#include "debug.h"

/* Language-dependent contents of a type. */

struct GTY (()) lang_type
{
  char dummy;
};

/* Language-dependent contents of a decl. */

struct GTY (()) lang_decl
{
  char dummy;
};

struct GTY(()) lang_identifier
{
  struct tree_identifier common;
};

/* The resulting tree type. */

union GTY ((desc ("TREE_CODE (&%h.generic) == IDENTIFIER_NODE"),
          chain_next ("CODE_CONTAINS_STRUCT (TREE_CODE (&%h.generic), "
          "TS_COMMON) ? ((union lang_tree_node *) TREE_CHAIN "
          "(&%h.generic)) : NULL"))) lang_tree_node
{
  union tree_node GTY ((tag ("0"), desc ("tree_node_structure (&%h)"))) generic;
  struct lang_identifier GTY ((tag ("1"))) identifier;
};

/* We don't use language_function. */

struct GTY (()) language_function
{
  int dummy;
};

/* Functions called directly by the generic backend. */

tree
convert (tree /* type */, tree /* expr */)
{
  gcc_unreachable ();
}

/**
* Эта функция запускается при подаче компилятору входного файла. Из неё мы
* начинаем наши действия
*/
static void
x_langhook_parse_file ()
{
  tree * main_params = new tree[1]; // Массив, содержащий параметры
  // У нас всего один параметр типа void
  main_params[0] = void_type_node;

  // Создаём тип функции "int (void)"
  tree main_fndecl_type = build_function_type_array (
    integer_type_node, // Тип возврата
    1,                 // Количество аргументов
    main_params);      // Массив с типами аргументов
  delete[] main_params;

  // Создаём объявление функции "int main(void);"
  tree main_fndecl = build_fn_decl(
    "main",            // Символьное имя функции
    main_fndecl_type); // Тип функции

  // Говорим что это объявление используется и его нельзя удалять
  DECL_PRESERVE_P (main_fndecl) = 1;
  // Функция main не должна иметь признака external
  DECL_EXTERNAL (main_fndecl)   = 0;

  // Создаём блок кода
  tree body                  = make_node(BLOCK);
  // Указываем охватывающее определение
  BLOCK_SUPERCONTEXT (body)  = main_fndecl;
  // Помечаем что попадаем в этот блок при входе в функцию
  DECL_INITIAL (main_fndecl) = body;

  // Создаём узел, в котором объявляется результат возврата функции
  tree resdecl = build_decl (
    UNKNOWN_LOCATION,   // Положение в исходном коде
    RESULT_DECL,        // Тип узла
    NULL_TREE,          // Узел с именем переменной
    integer_type_node); // Тип выражения

  // Наше объявление находится в контексте функции
  DECL_CONTEXT (resdecl)    = main_fndecl;
  // Помечаем объявленную переменную результатом
  DECL_RESULT (main_fndecl) = resdecl;

  //--------------------- Пример 1

  tree var_decl_1 = build_decl (
    UNKNOWN_LOCATION,        // Положение в исходном коде
    VAR_DECL,                // Тип узла
    get_identifier("var1"),  // Имя переменной
    integer_type_node);      // Тип переменной

  tree var_decl_2 = build_decl (
    UNKNOWN_LOCATION,        // Положение в исходном коде
    VAR_DECL,                // Тип узла
    get_identifier("var2"),  // Имя переменной
    boolean_type_node);      // Тип переменной

  tree var_decl_3 = build_decl (
    UNKNOWN_LOCATION,        // Положение в исходном коде
    VAR_DECL,                // Тип узла
    get_identifier("var3"),  // Имя переменной
    float_type_node);        // Тип переменной

  // Создаём цепочку объявлений
  DECL_CHAIN (var_decl_1) = var_decl_2;
  DECL_CHAIN (var_decl_2) = var_decl_3;

  // Помечаем что переменные нельзя удалять
  TREE_USED (var_decl_1)  = 1;
  TREE_USED (var_decl_2)  = 1;
  TREE_USED (var_decl_3)  = 1;

  // Список выражений
  tree stm_list = alloc_stmt_list ();

  // <-------------- Сюда будем добавлять код

  //--------------------- Пример 2

  tree const_decl_1 = build_int_cst (
    integer_type_node, // Тип константы
    10);               // Значение

  REAL_VALUE_TYPE r3; // Переменная с плавающим значением
  real_from_string ( &r3,    // Переменная, куда запишем
                     "3.4"); // Значение
  tree const_decl_3 = build_real (
      float_type_node, // Тип плавающего значения
      r3);             // Переменная, со значением

  // Говорим что объявления инициализируются константами
  DECL_INITIAL (var_decl_1) = const_decl_1;
  DECL_INITIAL (var_decl_3) = const_decl_3;

  // <-------------- Сюда будем добавлять код

  //--------------------- Пример 3

  // Создаём константу true
  tree const_decl_2 = build_int_cst_type (boolean_type_node, 1);

  // Присваиваем значение константы переменной
  tree bool_assign = fold_build2 (
    MODIFY_EXPR,    // Тип узла
    void_type_node, // Тип выражения
    var_decl_2,     // Цель присваивания
    const_decl_2);  // Присваиваемое значение

  // Добавляем выражение присваивания в список
  append_to_statement_list (
    bool_assign, // Добавляемое выражение
    &stm_list);  // Список выражений

  // <-------------- Сюда будем добавлять код

  //--------------------- Пример 4

  // Создаём константу
  tree const_decl_4 = build_int_cst_type (integer_type_node, 3);

  // Выражение умножения
  tree mult_tree = build2 (
    MULT_EXPR,         // Тип узла
    integer_type_node, // Тип выражения
    var_decl_1,        // Левый операнд
    const_decl_4);     // Правый операнд

  // Выражение умножения
  tree minus_tree = build2 (
    MINUS_EXPR,        // Тип узла
    integer_type_node, // Тип выражения
    mult_tree,         // Левый операнд
    const_decl_1);     // Правый операнд

  // Присваиваем значение выражения переменной
  tree var1_assign = fold_build2 (
    MODIFY_EXPR,    // Тип узла
    void_type_node, // Тип выражения
    var_decl_1,     // Цель присваивания
    minus_tree);    // Присваиваемое значение

  // Добавляем выражение присваитвания в список
  append_to_statement_list (
    var1_assign, // Добавляемое выражение
    &stm_list);  // Список выражений

  // <-------------- Сюда будем добавлять код

  //--------------------- Пример 5

  // Присваиваем значение переменной var_decl_1
  // переменной возврата
  tree ret_assign2 = fold_build2 (
    MODIFY_EXPR,     // Тип узла
    void_type_node,  // Тип выражения
    resdecl,         // Цель присваивания
    var_decl_1);     // Присваиваемое значение

  // Здесь создаём узел return
  tree return_stm2 = build1 (
    RETURN_EXPR,    // Тип узла
    void_type_node, // Тип выражения
    ret_assign2);   // Возвращаемое выражение

  // Добавляем выражение возврата в список
  append_to_statement_list (
    return_stm2, // Добавляемое выражение
    &stm_list);  // Список выражений

  // <-------------- Сюда будем добавлять код

  //--------------------- Пример 6

  // Создаём определение метки
  tree label_decl_1 = build_decl (
    UNKNOWN_LOCATION,           // Положение в исходном коде
    LABEL_DECL,                 // Тип узла
    get_identifier ("label_1"), // Узел с именем переменной
    void_type_node);            // Тип выражения

  // Указываем контекст для метки
  DECL_CONTEXT (label_decl_1) = main_fndecl;
  // Добавляем в цепочку определений
  DECL_CHAIN (var_decl_3) = label_decl_1;

  // Создаём выражение goto
  tree goto_1 = build1 (
    GOTO_EXPR,      // Тип узла
    void_type_node, // Тип выражения
    label_decl_1);  // Метка, на которую переходим

  // Строим саму метку
  tree label_expr_1 = build1 (
    LABEL_EXPR,     // Тип узла
    void_type_node, // Тип выражения
    label_decl_1);  // Объявление метки

  // Добавляем выражение перехода
  append_to_statement_list (
    goto_1,     // Добавляемое выражение
    &stm_list); // Список выражений

  // Сюда для эксперимента можно добавить выражений
  // которые мы перепрыгнем и не исполним

  // Добавляем выражение метки, в которую переходим
  append_to_statement_list (
    label_expr_1, // Добавляемое выражение
    &stm_list);   // Список выражений

  // <-------------- Сюда будем добавлять код

  //--------------------- Пример 7

  // Создаём определение метки
  tree true_label_decl = build_decl (
    UNKNOWN_LOCATION,              // Положение в исходном коде
    LABEL_DECL,                    // Тип узла
    get_identifier ("true_label"), // Узел с именем переменной
    void_type_node);               // Тип выражения

  // Указываем контекст для метки
  DECL_CONTEXT (true_label_decl) = main_fndecl;
  // Добавляем в цепочку определений
  DECL_CHAIN (label_decl_1) = true_label_decl;

  // Создаём выражение goto
  tree goto_true = build1 (
    GOTO_EXPR,        // Тип узла
    void_type_node,   // Тип выражения
    true_label_decl); // Метка, на которую переходим

  // Создаём определение метки
  tree false_label_decl = build_decl (
    UNKNOWN_LOCATION,               // Положение в исходном коде
    LABEL_DECL,                     // Тип узла
    get_identifier ("false_label"), // Узел с именем переменной
    void_type_node);                // Тип выражения

  // Указываем контекст для метки
  DECL_CONTEXT (false_label_decl) = main_fndecl;
  // Добавляем в цепочку определений
  DECL_CHAIN (true_label_decl) = false_label_decl;

  // Создаём выражение goto
  tree goto_false = build1 (
    GOTO_EXPR,         // Тип узла
    void_type_node,    // Тип выражения
    false_label_decl); // Метка, на которую переходим

  // Создаём определение метки
  tree endif_label_decl = build_decl (
    UNKNOWN_LOCATION,          // Положение в исходном коде
    LABEL_DECL,                // Тип узла
    get_identifier ("end_if"), // Узел с именем переменной
    void_type_node);           // Тип выражения

  // Указываем контекст для метки
  DECL_CONTEXT (endif_label_decl) = main_fndecl;
  // Добавляем в цепочку определений
  DECL_CHAIN (false_label_decl) = endif_label_decl;

  // Создаём выражение goto
  tree goto_endif = build1 (
    GOTO_EXPR,         // Тип узла
    void_type_node,    // Тип выражения
    endif_label_decl); // Метка, на которую переходим

  // Создаём узел с условным выражением
  tree cond_expr = build3 (
    COND_EXPR,      // Тип узла
    void_type_node, // Тип выражения
    var_decl_2,     // Выражение условия
    goto_true,      // Переход на истинную ветку
    goto_false);    // Переход на ложную ветку

  // Добавляем выражение сравнения
  append_to_statement_list (
    cond_expr,  // Добавляемое выражение
    &stm_list); // Список выражений

  // Строим метку начала истинного пути
  tree true_label_expr = build1 (
    LABEL_EXPR,       // Тип узла
    void_type_node,   // Тип выражения
    true_label_decl); // Объявление метки

  // Добавляем выражение метки, в которую переходим
  append_to_statement_list (
    true_label_expr, // Добавляемое выражение
    &stm_list);      // Список выражений

  // Здесь идёт код истинной ветки
  // Можно добавить узел BIND для создания
  // нового лексического блока

  // Добавляем выражение перехода на конец if'а
  append_to_statement_list (
    goto_endif, // Добавляемое выражение
    &stm_list); // Список выражений

  // Строим метку начала ложного пути
  tree false_label_expr = build1 (
    LABEL_EXPR,        // Тип узла
    void_type_node,    // Тип выражения
    false_label_decl); // Объявление метки

  // Добавляем выражение метки, в которую переходим
  append_to_statement_list (
    false_label_expr, // Добавляемое выражение
    &stm_list);       // Список выражений

  // Здесь идёт код ложной ветки
  // Можно добавить узел BIND для создания
  // нового лексического блока

  // Строим метку начала конца условного выражения.
  // Сюда мы прыгаем после окончания работы истиной ветки
  tree endif_label_expr = build1 (
    LABEL_EXPR,        // Тип узла
    void_type_node,    // Тип выражения
    endif_label_decl); // Объявление метки

  // Добавляем выражение метки, в которую переходим
  append_to_statement_list (
    endif_label_expr, // Добавляемое выражение
    &stm_list);       // Список выражений

  // <-------------- Сюда будем добавлять код

  //--------------------- Пример 8

  // Создаём определение метки
  tree loop_head_decl = build_decl (
    UNKNOWN_LOCATION,             // Положение в исходном коде
    LABEL_DECL,                   // Тип узла
    get_identifier ("loop_head"), // Узел с именем переменной
    void_type_node);              // Тип выражения

  // Указываем контекст для метки
  DECL_CONTEXT (loop_head_decl) = main_fndecl;
  // Добавляем в цепочку определений
  DECL_CHAIN (endif_label_decl) = loop_head_decl;

  // Создаём выражение goto
  tree goto_loop_head = build1 (
    GOTO_EXPR,       // Тип узла
    void_type_node,  // Тип выражения
    loop_head_decl); // Метка, на которую переходим

  // Создаём определение метки
  tree loop_exit_decl = build_decl (
    UNKNOWN_LOCATION,             // Положение в исходном коде
    LABEL_DECL,                   // Тип узла
    get_identifier ("loop_exit"), // Узел с именем переменной
    void_type_node);              // Тип выражения

  // Указываем контекст для метки
  DECL_CONTEXT (loop_exit_decl) = main_fndecl;
  // Добавляем в цепочку определений
  DECL_CHAIN (loop_head_decl) = loop_exit_decl;

  // Создаём переход на выход из цикла
  tree goto_loop_exit = build1 (
    GOTO_EXPR,       // Тип узла
    void_type_node,  // Тип выражения
    loop_exit_decl); // Метка, на которую переходим

  // Строим метку головы цикла
  tree loop_head_expr = build1 (
    LABEL_EXPR,      // Тип узла
    void_type_node,  // Тип выражения
    loop_head_decl); // Объявление метки

  // Добавляем выражение метки головы цикла
  append_to_statement_list (
    loop_head_expr, // Добавляемое выражение
    &stm_list);     // Список выражений

  // Здесь идёт код тела цикла
  // Можно добавить узел BIND для создания
  // нового лексического блока

  // Создаём узел с условным выражением
  tree cond_expr2 = build3 (
    COND_EXPR,       // Тип узла
    void_type_node,  // Тип выражения
    var_decl_2,      // Выражение условия
    goto_loop_head,  // Переход на истинную ветку
    goto_loop_exit); // Переход на ложную ветку

  // Добавляем выражение сравнения
  append_to_statement_list (
    cond_expr2, // Добавляемое выражение
    &stm_list); // Список выражений

  // Строим метку выхода из цикла
  tree loop_exit_expr = build1 (
    LABEL_EXPR,      // Тип узла
    void_type_node,  // Тип выражения
    loop_exit_decl); // Объявление метки

  // Добавляем выражение метки выхода из цикла
  append_to_statement_list (
    loop_exit_expr, // Добавляемое выражение
    &stm_list);     // Список выражений

  // <-------------- Конец примера

  // Создаём константу
  tree ret_const_decl = build_int_cst_type (integer_type_node, 0);
  // Присваиваем значение константы переменной возврата
  tree ret_assign     = fold_build2 (
    MODIFY_EXPR,     // Тип узла
    void_type_node,  // Тип выражения
    resdecl,         // Цель присваивания
    ret_const_decl); // Присваиваемое значение

  // Здесь создаём узел return
  tree return_stm = build1 (
    RETURN_EXPR,    // Тип узла
    void_type_node, // Тип выражения
    ret_assign);    // Возвращаемое выражение

  // Добавляем выражение возврата в список
  append_to_statement_list (
    return_stm, // Добавляемое выражение
    &stm_list); // Список выражений

  // Теперь соединяем блок кода со списком выражений
  tree bind = build3 (
    BIND_EXPR,      // Тип узла
    void_type_node, // Тип выражения
    var_decl_1,     // Список объявлений
    stm_list,       // Список выражений
    body);          // Охватывающий лексический блок

  // Задаём ссылку на выражение привязки к представлению
  DECL_SAVED_TREE (main_fndecl) = bind;

  // Преобразуем из GENERIC в GIMPLE
  gimplify_function_tree (main_fndecl);

  // Добавляем это в граф (видимо граф вызовов)
  cgraph_node::finalize_function (main_fndecl, true);
}

/* Language hooks. */

static bool
x_langhook_init (void)
{
  build_common_tree_nodes (false);

  build_common_builtin_nodes ();

  return true;
}

static tree
x_langhook_type_for_mode (enum machine_mode mode, int unsignedp)
{
  if (mode == TYPE_MODE (float_type_node))
    return float_type_node;

  if (mode == TYPE_MODE (double_type_node))
    return double_type_node;

  if (mode == TYPE_MODE (intQI_type_node))
    return unsignedp ? unsigned_intQI_type_node : intQI_type_node;
  if (mode == TYPE_MODE (intHI_type_node))
    return unsignedp ? unsigned_intHI_type_node : intHI_type_node;
  if (mode == TYPE_MODE (intSI_type_node))
    return unsignedp ? unsigned_intSI_type_node : intSI_type_node;
  if (mode == TYPE_MODE (intDI_type_node))
    return unsignedp ? unsigned_intDI_type_node : intDI_type_node;
  if (mode == TYPE_MODE (intTI_type_node))
    return unsignedp ? unsigned_intTI_type_node : intTI_type_node;

  if (mode == TYPE_MODE (integer_type_node))
    return unsignedp ? unsigned_type_node : integer_type_node;

  if (mode == TYPE_MODE (long_integer_type_node))
    return unsignedp ? long_unsigned_type_node : long_integer_type_node;

  if (mode == TYPE_MODE (long_long_integer_type_node))
    return unsignedp ? long_long_unsigned_type_node
        : long_long_integer_type_node;

  if (COMPLEX_MODE_P (mode))
  {
    if (mode == TYPE_MODE (complex_float_type_node))
      return complex_float_type_node;
    if (mode == TYPE_MODE (complex_double_type_node))
      return complex_double_type_node;
    if (mode == TYPE_MODE (complex_long_double_type_node))
      return complex_long_double_type_node;
    if (mode == TYPE_MODE (complex_integer_type_node) && !unsignedp)
      return complex_integer_type_node;
  }

  /* gcc_unreachable */
  return NULL;
}

/* Record a builtin function. We just ignore builtin functions. */

static tree
x_langhook_builtin_function (tree decl)
{
  return decl;
}

static bool
x_langhook_global_bindings_p (void)
{
  gcc_unreachable ();
  return true;
}

static tree
x_langhook_pushdecl (tree decl ATTRIBUTE_UNUSED)
{
  gcc_unreachable ();
}

static tree
x_langhook_getdecls (void)
{
  return NULL;
}

#undef LANG_HOOKS_NAME
#define LANG_HOOKS_NAME "X"

#undef LANG_HOOKS_INIT
#define LANG_HOOKS_INIT x_langhook_init

#undef LANG_HOOKS_PARSE_FILE
#define LANG_HOOKS_PARSE_FILE x_langhook_parse_file

#undef LANG_HOOKS_TYPE_FOR_MODE
#define LANG_HOOKS_TYPE_FOR_MODE x_langhook_type_for_mode

#undef LANG_HOOKS_BUILTIN_FUNCTION
#define LANG_HOOKS_BUILTIN_FUNCTION x_langhook_builtin_function

#undef LANG_HOOKS_GLOBAL_BINDINGS_P
#define LANG_HOOKS_GLOBAL_BINDINGS_P x_langhook_global_bindings_p

#undef LANG_HOOKS_PUSHDECL
#define LANG_HOOKS_PUSHDECL x_langhook_pushdecl

#undef LANG_HOOKS_GETDECLS
#define LANG_HOOKS_GETDECLS x_langhook_getdecls

struct lang_hooks lang_hooks = LANG_HOOKS_INITIALIZER;

#include "gt-x-x1.h"
#include "gtype-x.h"
